from constants import HOST, API_BASE_URL, CSV_DATA_PATH, SCHEMA_PATH
from database import Database
from functions import log
import xml.etree.ElementTree as ET
import requests


class API:
    """
    This class contains methods that perform API calls
    """

    def __init__(self, database:Database):

        self.db = database
        self.php_sess_id = self.db.select("SELECT value FROM config WHERE key = 'PHPSESSID'")
        if self.php_sess_id is None: self.php_sess_id = ''


    def api_call(self,
        url:str,
        index:int=2,
        headers:dict={},
        file_path:str='',
        try_to_login:bool=True) -> ET.ElementTree:
        """
        This function performs an API call and returns the response
        in the form of an ElementTree object

        If login to Analytics is required, this function will attempt to login
        to Analytics once, and then redo the API call

        If login to Analytics fails, or the API call fails for any reason,
        this function returns a garbage ElementTree object
        """

        hds = { # Headers that all API calls require
            'Host':             f'{HOST}.smtresearch.ca',
            'User-Agent':       'Mozilla/5.0',
            'Accept':           'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
            'Accept-Language':  'en-US,en;q=0.8',
            'Connection':       'keep-alive'
        }
        hds.update(headers) # Some API calls require extra headers
        cks = {'PHPSESSID': self.php_sess_id}

        try:
            if   index == 1: response = requests.get(url, headers=hds)
            elif index == 2: response = requests.get(url, headers=hds, cookies=cks)
            elif index == 3: response = requests.put(url, headers=hds, cookies=cks, data=open(file_path, 'rb').read())

            # root = ElementTree object
            root = ET.fromstring(response.content)

            # If login is not required
            if root.text != 'login required': return root

            log(f'{HOST} login required [PHPSESSID {self.php_sess_id}]')

            if try_to_login and self.login():
                # login() will update self.php_sess_id
                # Redo API call with updated self.php_sess_id
                # Make sure try_to_login=False to avoid an infinite loop
                return self.api_call(url, index, headers, file_path, False)
            
            # If try_to_login=False or login()=False
            return ET.fromstring('<API_failed></API_failed>')

        except Exception as e:
            log(f'API call failed: {url} | {e}')
            return ET.fromstring('<API_failed></API_failed>')


    def login(self, username=None, password=None) -> bool:

        if username is None:
            username = self.db.select("SELECT value FROM config WHERE key = 'analyticsUsername'")
        
        if password is None:
            password = self.db.select("SELECT value FROM config WHERE key = 'analyticsPassword'")

        log(f'Logging into {HOST}...')

        url = f'{API_BASE_URL}login&user_username={username}&user_password={password}'
        root = self.api_call(url, index=1, try_to_login=False)

        result = root.find('login')
        if result is None or result.text != 'success':
            log('Login failed')
            return False

        # Login successful -> update self.php_sess_id
        self.php_sess_id = root.find('PHPSESSID').text

        self.db.execute(f"UPDATE config SET value = '{self.php_sess_id}' WHERE key = 'PHPSESSID'")
        log(f'Login successful [PHPSESSID {self.php_sess_id}]')

        return True


    def prompt_user_login(self) -> bool:

        try:
            while True:
                username = input(f'\nEnter {HOST} username: ')
                password = input(f'Enter {HOST} password: ')

                if self.login(username, password):
                    self.db.execute(f"UPDATE config SET value = '{username}' WHERE key = 'analyticsUsername'")
                    self.db.execute(f"UPDATE config SET value = '{password}' WHERE key = 'analyticsPassword'")
                    return True

                try_again = input('Try logging in again? [y/n] ')
                while try_again not in ['y','Y','n','N']:
                    try_again = input('Invalid input. Try logging in again? [y/n] ')

                if try_again in ['n','N']: return False
        except KeyboardInterrupt: return False # CTRL+C to exit


    def logout(self):

        log(f'Logging out of {HOST}...')

        url = f'{API_BASE_URL}logout'
        root = self.api_call(url, try_to_login=False)

        result = root.find('logout')
        if result is None or result.text != 'success':
            log('Logout failed')
        else: log('Logout successful')


    def commission(self, gw_id:int, project_id:int) -> str:
        
        log(f'Commissioning TiG... [gateway ID {gw_id} | project ID {project_id}]')
        
        url = f'{API_BASE_URL}commission&gwID={gw_id}&projectID={project_id}'
        root = self.api_call(url)

        gw_key = root.find('commission/gwKey')
        if gw_key is None or gw_key.text is None:
            log('Commission failed')
            return False

        gw_key = gw_key.text
        log(f'TiG commissioned [gateway key {gw_key}]')

        return gw_key


    def decommission(self, gw_id:int):

        log(f'Decommissioning TiG... [gateway ID {gw_id}]')

        url = f'{API_BASE_URL}decommission&gwID={gw_id}'
        root = self.api_call(url)

        if root.find('decommission') is None:
            log('Decommission failed')     
        else: log('TiG decommissioned')


    def serialize(self, gw_name:str) -> str:

        log(f'Serializing TiG... [gateway name {gw_name}]')
        
        url = f'{API_BASE_URL}serialize&gw_name={gw_name}'
        root = self.api_call(url)
        
        gw_id = root.find('gateway/gwID')
        if gw_id is None:
            log('Serialize failed')
            return False
        
        gw_id = int(gw_id.text)
        log(f'TiG serialized [gateway ID {gw_id}]')

        return gw_id


    def unserialize(self, gw_id:int):

        log(f'Unserializing TiG... [gateway ID {gw_id}]')

        url = f'{API_BASE_URL}unserialize&gwID={gw_id}'
        root = self.api_call(url)

        if root.find('gateway') is None:
            log('Unserialize failed')
        else: log('TiG unserialized')


    def get_nodes(self, job_id:int):
        
        url = f'{API_BASE_URL}listNode&jobID={job_id}'
        root = self.api_call(url)

        nodes = root.find('nodes')
        if nodes is None:
            log(f'Unable to get node list from {HOST} [job ID {job_id}]')
            return []

        return nodes


    def get_sensors(self, node_id:int):

        url = f'{API_BASE_URL}listSensor&nodeID={node_id}'
        root = self.api_call(url)

        sensors = root.find('sensors')
        if sensors is None:
            log(f'Unable to get sensors from {HOST} [node ID {node_id}]')
            return []

        return sensors


    def upload_data(self, csv_string:str) -> bool:

        print('Uploading data...')

        try:
            with open(CSV_DATA_PATH, 'w') as f: f.write(csv_string)
        except Exception as e:
            log(f'Unable to open {CSV_DATA_PATH} | {e}')
            return False

        url = f'{API_BASE_URL}uploadData'
        headers = {'Content-Type': 'text/csv'}
        root = self.api_call(url, index=3, headers=headers, file_path=CSV_DATA_PATH)

        if root.find('success') is None:
            log('Upload failed')
            return False

        print('Upload successful')
        return True


    def exchange_schema(self) -> bool:

        log(f'Exchanging schema with {HOST}...')
        
        try:
            local_schema = self.create_schema()
            with open(SCHEMA_PATH, 'w') as f: f.write(local_schema)
        except Exception as e:
            log(f'Unable to open {SCHEMA_PATH} | {e}')
            return False
        
        print(f'Schema created. Syncing schema with {HOST}...')

        url = f'{API_BASE_URL}syncSchema'
        headers = {'Content-Type': 'text/xml', 'Content-Length': f'{len(local_schema)}'}
        analytics_schema = self.api_call(url, index=3, headers=headers, file_path=SCHEMA_PATH)

        return self.parse_schema(analytics_schema)


    def create_schema(self) -> str:

        print('Creating schema...')

        # Root of tree: schema
        schema_element = ET.Element('schema', version="2")

        # First generation: gateway
        gw_id   = self.db.select('SELECT gwID FROM gateway')
        gw_name = self.db.select('SELECT name FROM gateway')

        gateway_element = ET.SubElement(schema_element, 'gateway', {
        'gwID': f'{gw_id}', 'name': f'{gw_name}',
        'last-modified': '1990-01-01 12:00:00'})

        # Second generation: job
        job_id = self.db.select("SELECT jobID FROM job")

        job_element = ET.SubElement(gateway_element, 'job', {
        'gid':           f'{job_id}',
        'address':       '',
        'contactName':   '',
        'contactPhone':  '',
        'created':       '',
        'last-modified': '1990-01-01 12:00:00'})

        # Third generation: node
        for node in self.db.select('SELECT * FROM node', fetch_one=False):
            row_id  = node[0]
            node_id = node[1]
            phy_id  = node[2]

            if node_id is None or node_id <= 0:
                node_attributes = {'id': f'{row_id}'}
            else: node_attributes = {'gid': f'{node_id}'}

            node_attributes.update({
            'phyID':         f'{phy_id}',
            'name':          f'{node[3]}',
            'created':       f'{node[4]}',
            'last-modified': f'{node[5]}'})

            node_element = ET.SubElement(job_element, 'node', node_attributes)

            # Fourth generation: sensor     
            for sensor in self.db.select(f'SELECT * FROM sensor WHERE phyID = {phy_id}', fetch_one=False):
                row_id    = sensor[0]
                sensor_id = sensor[1]

                if sensor_id is None or sensor_id <= 0:
                    sensor_attributes = {'id': f'{row_id}'}
                else: sensor_attributes = {'gid': f'{sensor_id}'}

                sensor_attributes.update({
                'name':           f'{sensor[2]}',
                'type':           f'{sensor[3]}',
                'input':          f'{sensor[4]}',
                'comments':       f'{sensor[5]}',
                'wsID':           f'{sensor[6]}',
                'sensorTypeID':   f'{sensor[7]}',
                'sensorTypeName': f'{sensor[8]}',
                'offset':         f'{sensor[9]}',
                'custom1':        f'{sensor[10]}',
                'custom2':        f'{sensor[11]}',
                'custom3':        f'{sensor[12]}',
                'created':        f'{sensor[13]}',
                'last-modified':  f'{sensor[14]}'})

                sensor_element = ET.SubElement(node_element, 'sensor', sensor_attributes)

        return ET.tostring(schema_element).decode()


    def parse_schema(self, analytics_schema:ET.ElementTree) -> bool:

        job = analytics_schema.find('gateway/job')
        if job is None:
            log(f'Unable to parse {HOST} schema')
            return False
        
        log('Parsing schema...')

        # Variables used for log messages
        nodes_to_insert = nodes_inserted = 0
        nodes_to_update = nodes_updated  = 0
        
        sensors_to_insert = sensors_inserted = 0
        sensors_to_update = sensors_updated  = 0
        
        ############################################# LOOP THROUGH NODES #############################################
        for node in job:

            row = []; missing_attributes = []

            for attribute in ['gid', 'phyID', 'name', 'created', 'last-modified']:
                if attribute in node.attrib: row.append(node.attrib[attribute])
                else: missing_attributes.append(attribute)

            if missing_attributes:
                log(f'Schema missing node attributes: {", ".join(missing_attributes)}')
                continue # Skip this node

            phy_id = node.attrib['phyID']
            
            if 'id' in node.attrib: # If this is a new node
                row.insert(0, node.attrib['id'])

                query = f"INSERT OR REPLACE INTO node (id, nodeID, phyID, name, created, modified) VALUES {tuple(row)}"
                nodes_inserted += self.db.execute(query)
                nodes_to_insert += 1
            
            else: # If this is not a new node
                query = f"UPDATE node SET (nodeID, phyID, name, created, modified) = {tuple(row)} WHERE phyID = {phy_id}"
                
                tig_mod = self.db.select(f'SELECT modified FROM node WHERE phyID = {phy_id}')
                anl_mod = node.attrib['last-modified']
                
                if tig_mod is not None and anl_mod > tig_mod: # If Analytics modified date is more recent
                    nodes_updated += self.db.execute(query)
                    nodes_to_update += 1
            
            ############################################ LOOP THROUGH SENSORS ############################################
            for sensor in node:
                
                row = []; missing_attributes = []

                for attribute in ['gid','name','type','input','comments','wsID','offset','custom1','custom2','custom3','created','last-modified']:
                    if attribute in sensor.attrib: row.append(sensor.attrib[attribute])
                    else: missing_attributes.append(attribute)

                if missing_attributes:
                    log(f'Schema missing sensor attributes: {", ".join(missing_attributes)}')
                    continue # Skip this sensor
                
                row.append(phy_id)
                
                if 'id' in sensor.attrib: # If this is a new sensor
                    row.insert(0, sensor.attrib['id'])

                    query = f"INSERT OR REPLACE INTO sensor (id, sensorID, name, type, input, comments, wsID, offset, custom1, custom2, custom3, created, modified, phyID) VALUES {tuple(row)}"
                    sensors_inserted += self.db.execute(query)
                    sensors_to_insert += 1

                else: # If this is not a new sensor
                    sensor_input = sensor.attrib['input']
                    query = f"UPDATE sensor SET (sensorID, name, type, input, comments, wsID, offset, custom1, custom2, custom3, created, modified, phyID) = {tuple(row)} WHERE input = {sensor_input} AND phyID = {phy_id}"
                    
                    tig_mod = self.db.select(f'SELECT modified FROM sensor WHERE input = {sensor_input} AND phyID = {phy_id}')
                    anl_mod = sensor.attrib['last-modified']

                    if tig_mod is not None and anl_mod > tig_mod: # If Analytics modified date is more recent
                        sensors_updated += self.db.execute(query)
                        sensors_to_update += 1
        
        log(f'{nodes_inserted}/{nodes_to_insert} nodes and {sensors_inserted}/{sensors_to_insert} sensors inserted in database')
        log(f'{nodes_updated}/{nodes_to_update} nodes and {sensors_updated}/{sensors_to_update} sensors updated in database')

        return True
