from constants import B_LED_ON, B_LED_OFF, ZMQ_SOCKET
from functions import get_service_pid, log
from usbhub import USBHub
from sync import Sync
import signal
import time
import zmq
import os


class BiGBerry:
    """
    This is the main class for running BiGBerry
    
    It contains the main infinite while loop that will run until:
        -the user manually ends this script
        -the user manually shuts down / restarts the RPi 
        -the RPi restarts automatically at midnight

    This class creates two important objects:
        1. A USBHub object (from usbhub.py)
        2. A Sync object (from sync.py)

    The USBHub object is mainly responsible for receiving sensor data
    sent to USB devices connected to the TiG, and saving the data
    to the local sqlite database on the TiG

    The Sync object is responsible for periodically syncing the data
    in the local sqlite database on the TiG
    with SMT's Analytics server database

    This class also catches and handles all kill signals
    that are executed while BiGBerry is running
    """

    def __init__(self):
        # Variable that will remain True until a KeyboardInterrupt (CTRL+C) is received
        # Note: KeyboardInterrupt (CTRL+C) is equivalent to kill -2 <PID of this script>
        self.running = True

        # For communication with ifc_main.py
        self.socket = zmq.Context().socket(zmq.PAIR)
        self.socket.connect(ZMQ_SOCKET)

        # For creating/deleting Wimote and CANKey objects, communcating with them,
        # and saving their incoming data to the local sqlite database on the TiG
        self.usb = USBHub(self.socket)

        # For syncing data in local sqlite database on the TiG
        # to the SMT Analytics server database
        self.sync = Sync(self.usb, self.socket)

        # This allows kill signals to be handled by the signal_handler method of this class
        # DO NOT include signals SIGINT (kill -2) or SIGKILL (kill -9) in this list
        for sig in [
            signal.SIGHUP,  # kill -1  -> bigberry sync
            signal.SIGQUIT, # kill -3  -> bigberry setST
            signal.SIGILL,  # kill -4  -> bigberry setPAN
            signal.SIGTRAP, # kill -5  -> bigberry reTX
            signal.SIGABRT, # kill -6  -> bigberry erase
            signal.SIGBUS,  # kill -7  -> bigberry updateUSB
            signal.SIGFPE,  # kill -8  -> bigberry writeUSBinfo
            signal.SIGUSR1  # kill -10 -> bigberry logUSB
        ]: signal.signal(sig, self.signal_handler)


    def begin(self):

        # Turn on B LED on TiG ifc
        self.socket.send_pyobj(B_LED_ON)

        # Start threads
        self.usb.thread.start()
        self.sync.thread.start()
        
        # Check for any USB devices connected to RPi
        self.usb.update_usb_lists()

        # Begin infinite while loop
        try:
            while self.running: time.sleep(1)
        # Note: KeyboardInterrupt (CTRL+C) is equivalent to kill -2 <PID of this script>
        except KeyboardInterrupt: self.running = False

    
    def end(self):

        log(f'Ending bb_main.py [PID {current_pid}]...')

        self.usb.stop_devices()

        # Stop threads
        self.usb.running = False
        self.usb.thread.join()

        self.sync.running = False
        self.sync.thread.join()

        self.usb.db.close()
        
        # Turn off B LED on TiG ifc
        self.socket.send_pyobj(B_LED_OFF)

    
    def signal_handler(self, k:int, frame):
        """
        This method is called whenever a kill signal <k> is executed
        For example: kill -k <PID of this script>
        """

        # 1 sync | Used by RaspAp web UI and ifc_main.py
        if k == 1: self.sync.force_sync()

        # 3 setST, 4 setPAN, 5 reTX, 6 erase | Used by RaspAp web UI
        elif 3 <= k <= 6: self.usb.wimote_write(k)

        # 7 updateUSB | Used by crontab and /etc/udev/rules.d/88-bigberry.rules
        elif k == 7: self.usb.update_usb_lists()

        # 8 writeUSBinfo | Used by RaspAp web UI
        elif k == 8: self.usb.write_usb_info()

        # 10 logUSB | Used for testing
        elif k == 10: self.usb.log_usb_info()




if __name__ == '__main__':
    """
    It is important that only one instance of this script is running at a time
    This is checked by comparing the PID of this script with the PID of bigberry.service

    If this script was run by starting bigberry.serivce, the two PIDs will be the same,
    and this script will start BiGBerry

    Otherwise, the two PIDs will be different, and this script will do nothing

    bigberry.service can only be started once, so running this script by starting bigberry.service
    will ensure that only one instance of this script can ever be running at a time

    However, this means that the ONLY way to run this script is by starting bigberry.service
    To bypass this, remove the if-else statement below
    """

    current_pid = os.getpid()
    service_pid = get_service_pid('bigberry.service')

    log(f'############## RUNNING bb_main.py [PID {current_pid}] ##############')

    if current_pid == service_pid:
        bigberry = BiGBerry()
        bigberry.begin()
        # Wait for KeyboardInterrupt (CTRL+C) or kill -2 <PID of this script>
        bigberry.end()
    else:
        log(f'ERROR! bb_main.py PID does not match bigberry.service PID [{service_pid}]')

    log(f'############## END OF bb_main.py [PID {current_pid}] ##############')
    