from datetime import datetime
from functions import log
import threading
import usb.core
import usb.util
import queue
import time


class CANKey:
    """
    Instances of this class are created in the 'update_usb_lists' method
    of the USBHub class in usbhub.py

    The list 'cankeys' in USBHub will contain one instance of this class
    for each cankey connected to the TiG

    The 'read' method of this class is from the perspective of the TiG
    It will wait for messages sent from the cankey to the TiG
    """

    def __init__(self, dev, data_queue:queue.Queue):

        # pyUSB device object
        self.dev = dev

        if self.dev.is_kernel_driver_active(0): 
            self.dev.detach_kernel_driver(0)

        self.address = dev.address # integer (unique)
        self.idProduct = dev.idProduct # 0xff0f
        self.ep_in, self.ep_out = dev[0][(0,0)]

        # Data received by this cankey object gets put into this queue
        # This is the same queue used by all Wimote and CANKey objects
        # in the 'wimote' and 'cankey' lists in USBHub
        self.data_queue = data_queue

        # Dictionary of A3's connected to this Wimote
        # key: node ID
        # value: Number of readings received since last sync
        self.nodes = {}

        self.thread = threading.Thread(target=self.read)
        self.running = True
        self.thread.start()
        
        log(f'[CANKey {self.address}] - CONNECTED')


    # Thread
    def read(self):
        
        while self.running:
            
            # Try to receive a message from this cankey device
            # If there are no messages, this will raise a timeout error
            try: rx = self.dev.read(self.ep_in.bEndpointAddress, 64)
            except usb.core.USBError as e:
                if e.args[0] == 110: continue # Ignore timeout error
                else: self.stop(err_msg=f'READING ERROR: {e}')

            #
            # Message received -> proceed
            #

            # First number in message is the command
            cmd = rx[5]

            if cmd == 1:
                # CANKey is receiving sensor data
                
                node_id = (rx[2] & 0xff) << 8 | (rx[3] & 0xff)
                if node_id >= 65535: continue # Ignore messages with node_id >= 65535

                # If this is the first message received from this node, add it to self.nodes
                if node_id not in self.nodes:
                    self.nodes[node_id] = 0
                    log(f'[CANKey {self.address}] - Receiving readings from node {node_id}')

                ts = int(time.time()) # Timestamp is saved in database in UTC time
                sensor_val = (rx[6] & 0xff) << 24 | (rx[7] & 0xff) << 16 | (rx[8] & 0xff) << 8 | (rx[9] & 0xff)

                sensor_input = rx[10] + 1
                if sensor_input < 256:
                    self.data_queue.put_nowait((node_id, sensor_input, ts, sensor_val, 0))
                    self.nodes[node_id] += 1

            elif cmd == 4:
                fw_version = (rx[6] << 8) | rx[7]
                hw_version = rx[8]
                bl_version = rx[9]

                log(f'[CANKey {self.address}] Device: [{node_id}] FW: [{fw_version}] HW: [{hw_version}] BL: [{bl_version}]')


    def stop(self, err_msg=''):

        if self.running == False: return

        msg = f'[CANKey {self.address}] - STOPPING'
        if err_msg: msg += f' | {err_msg}'
        log(msg)

        self.log()
        self.running = False
        self.stop_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")


    def log(self, reset_counts=False):

        if not self.running:
            log(f'[CANKey {self.address}] - STOPPED RUNNING AT {self.stop_time}')
            return

        for node_id in self.nodes:
            count = self.nodes[node_id] # Number of readings received since last sync
            if count == 0: continue

            log(f'[CANKey {self.address} - node {node_id}] - Readings since last sync: {count}')
            if reset_counts: self.nodes[node_id] = 0
