from datetime import datetime, timezone
from constants import TMATE_CONF_PATH
from functions import shell_cmd, log
from database import Database
from sync import Sync
from api import API
import argparse


class Commission:

    def __init__(self, database:Database):
        # Connection to local sqlite database
        self.db = database

        # For making API calls
        self.api = API(self.db)
        
        # For syncing to Analytics
        self.sync = Sync(None, None)

        # Current hostname of RPi
        self.hostname = shell_cmd('cat /etc/hostname')

        # Get arguments (if any)
        parser = argparse.ArgumentParser()
        parser.add_argument('--gwName')
        parser.add_argument('--projectID')
        parser.add_argument('--jobID')
        args = parser.parse_args()

        self.gw_name    = args.gwName
        self.project_id = args.projectID
        self.job_id     = args.jobID

        self.commission()


    def commission(self):

        # If this script was run with any missing arguments, get them from user
        if self.gw_name is None or self.project_id is None or self.job_id is None:
            # If user cancels with CTRL+C, cancel commission
            if self.prompt_user_commission() == False: return

            # If user is unable to login to Analytics, cancel commission
            if self.api.prompt_user_login() == False: return


        gw_id = self.api.serialize(self.gw_name)
        if gw_id == False: return

        self.gw_key = self.api.commission(gw_id, self.project_id)
        if self.gw_key == False: return

        self.db.execute(f"UPDATE config SET value = {self.project_id} WHERE key = 'projectID'")
        self.db.execute(f"UPDATE config SET value = {self.project_id} WHERE key = 'pan'")
        self.db.execute( "UPDATE config SET value = 1 WHERE key = 'commissioned'")
        
        utc_dt = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        self.db.execute(f"UPDATE job SET (jobID, modified) = {(self.job_id, utc_dt)}")
        self.db.execute(f"UPDATE gateway SET (gwID, gwKey, modified) = {(gw_id, self.gw_key, utc_dt)}")
        self.db.execute(f"INSERT INTO node (phyID, name, modified) VALUES {(gw_id, self.gw_name, utc_dt)}")
        
        log('Config, job, gateway, and node tables updated')

        self.insert_nodes_sensors()
        # Sync to Analytics
        self.sync.sync_data(self.db, self.api)
        
        # If hostname = 'raspberrypi', assume this is the first time the TiG is being commissioned
        # Therefore, .tmate.conf and hostname need to be updated

        # .tmate.conf is only updated when the TiG is commissioned for the first time
        # to ensure the gateway key in .tmate.conf always matches the gateway key
        # in Analytics inventory, even after any subsequent commissions

        if self.hostname == 'raspberrypi':
            # Change owner and group of /home/pi/.tmate.conf and /etc/hostname to 'pi'
            # so this script has permission to open and edit those files
            shell_cmd('sudo bigberry setUser pi')

            self.update_hostname_file()
            self.update_tmate_file()

            # Change owner and group of /home/pi/.tmate.conf and /etc/hostname back to 'root'
            shell_cmd('sudo bigberry setUser root')
        

    def prompt_user_commission(self):
        """
        This method will update self.gw_name, self.project_id, and self.job_id
        by prompting the user for this information

        The user will only be prompted to change the gateway name of the TiG
        if the hostname of the RPi is still set to the default 'raspberrypi'

        After commissioning, the hostname of the RPi will be updated to the gateway name of the TiG,
        so the chosen gateway name must obey the rules for the hostname of the RPi

        The gateway name should be 'tigXXXX' where XXXX is a 4-digit number
        example gateway name: 'tig1234'
        """

        try:
            self.gw_name = self.hostname
            while self.hostname == 'raspberrypi':

                # Prompt user for new gateway name
                valid_name = False
                while not valid_name:
                    self.gw_name = input('\nEnter TiG gateway name: ')

                    if self.gw_name == 'raspberrypi':
                        print('Name cannot be "raspberrypi"')
                    elif len(self.gw_name) < 1:
                        print('Must enter a name')
                    elif len(self.gw_name) > 253: 
                        print('Name is too long (Max 253 characters)')
                    elif self.gw_name[0] == '-':
                        print('Name must begin with a letter or digit')
                    else:
                        valid_name = True
                        for char in self.gw_name:
                            if char.isdigit() or char.isalpha() or char == '-': continue
                            print(f'Name contains invalid character: {char}')
                            print('Valid characters are:')
                            print('  letters (a..z / A..Z)')
                            print('  numbers (0...9)')
                            print('  hyphens (-)')
                            valid_name = False
                            break

                print(f'\nTiG gateway name = {self.gw_name}')

                confirm = input('Is this correct? [y/n] ')
                while confirm not in ['y','Y','n','N']:
                    confirm = input('Invalid input. Is this correct? [y/n] ')

                if confirm in ['y','Y']: break
            ########################################################################################

            ######################### Update self.project_id and self.job_id #######################
            while True:

                self.project_id = input('\nEnter project ID: ')
                while not (self.project_id.isnumeric() and int(self.project_id) > 0):
                    print('Invalid input. Must be an integer greater than 0')
                    self.project_id = input('Enter project ID: ')

                self.job_id = input('Enter job ID: ')
                while not (self.job_id.isnumeric() and int(self.job_id) > 0):
                    print('Invalid input. Must be an integer greater than 0')
                    self.job_id = input('Enter job ID: ')

                print(f'\nProject ID = {self.project_id}\nJob ID = {self.job_id}')

                confirm = input('Is this correct? [y/n] ')
                while confirm not in ['y','Y','n','N']:
                    confirm = input('Invalid input. Is this correct? [y/n] ')

                if confirm in ['y','Y']: break
            ########################################################################################
        except KeyboardInterrupt: return False # CTRL+C to exit


    def insert_nodes_sensors(self):
        """Insert all nodes and sensors that exist for this job into database"""

        log(f'Inserting nodes and sensors [job ID {self.job_id}]...')

        node_rows = []; sensor_rows = []

        # Loop through nodes in api response
        for node in self.api.get_nodes(self.job_id):

            row = []; missing_attributes = []

            for attribute in ['nodeID', 'phyID', 'name', 'created', 'modified']:
                attr = node.find(attribute)
                if attr is None: missing_attributes.append(attribute)
                else: row.append(attr.text)

            if missing_attributes:
                print(f'xml tree missing node attributes: {", ".join(missing_attributes)}')
                continue # Skip this node

            node_id = row[0]
            phy_id  = row[1]

            node_rows.append(tuple(row))

            # Loop through sensors in api response
            for sensor in self.api.get_sensors(node_id):

                row = []; missing_attributes = []

                for attribute in ['sensorID','name','type','input','comments','wsID','sensorTypeID','sensorTypeName','offset','custom1','custom2','custom3','created','modified']:
                    attr = sensor.find(attribute)
                    if attr is None: missing_attributes.append(attribute)
                    else: row.append(attr.text)

                if missing_attributes:
                    print(f'xml tree missing sensor attributes: {", ".join(missing_attributes)}')
                    continue # Skip this sensor

                row.append(-1) # tempCompSensorID
                row.append(phy_id)
                
                sensor_rows.append(tuple(row))

        if node_rows:
            nodes_inserted = self.db.execute(
            """INSERT OR REPLACE INTO node (nodeID, phyID, name, created, modified) 
            VALUES (?, ?, ?, ?, ?)""", parameters=node_rows)
        else: nodes_inserted = 0

        if sensor_rows:
            sensors_inserted = self.db.execute(
            """INSERT OR REPLACE INTO sensor
            (sensorID, name, type, input, comments, wsID, sensorTypeID, sensorTypeName, 
            offset, custom1, custom2, custom3, created, modified, tempCompSensorID, phyID)
            VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)""", parameters=sensor_rows)
        else: sensors_inserted = 0
        
        log(f'{nodes_inserted}/{len(node_rows)} nodes and {sensors_inserted}/{len(sensor_rows)} sensors inserted into database')


    def update_hostname_file(self):

        try:
            with open('/etc/hostname', 'w') as f: f.write(self.gw_name)
            self.db.execute(f"UPDATE gateway SET name = '{self.gw_name}'")
            log(f'Hostname set to {self.gw_name} | Restart RPi')
        except Exception as e:
            log(f'Unable to open /etc/hostname | {e}')


    def update_tmate_file(self):

        try:
            with open(TMATE_CONF_PATH, 'w') as f:
                f.write('set tmate-api-key "tmk-9UqKt3c0UWA56SFWrquftaXXUK"\n')
                f.write(f'set tmate-session-name "{self.gw_name}-{self.gw_key}"')
            log('Updated .tmate.conf file')
        except Exception as e:
            log(f'Unable to open {TMATE_CONF_PATH} | {e}')




class Decommission:

    def __init__(self, database:Database):
        # Connection to local sqlite database
        self.db = database

        # For making API calls
        self.api = API(self.db)

        parser = argparse.ArgumentParser()
        parser.add_argument('--unserialize')
        args = parser.parse_args()

        self.unserialize = args.unserialize

        self.decommission()


    def decommission(self):

        # If script was run with any missing arguments, get them from user
        if self.unserialize is None:
            # If user cancels with CTRL+C, cancel commission
            if self.prompt_user_decommission() == False: return

            # If user is unable to login to Analytics, cancel decommission
            if self.api.prompt_user_login() == False: return


        gw_id = self.db.select("SELECT gwID FROM gateway")
        if self.api.decommission(gw_id) == False: return
    
        self.db.execute("DELETE FROM data")
        self.db.execute("DELETE FROM node")
        self.db.execute("DELETE FROM sensor")

        self.db.execute("UPDATE config SET value = '' WHERE key = 'gwKey'")
        self.db.execute("UPDATE config SET value = 0  WHERE key = 'commissioned'")

        utc_dt = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        self.db.execute(f"UPDATE gateway SET (gwKey, modified) = {('', utc_dt)}")

        log('Data, node, and sensor tables cleared')
        log('Config and gateway tables updated')

        if self.unserialize: self.api.unserialize(gw_id)
        else: log('Tig not unserialized')

        self.api.logout()

    
    def prompt_user_decommission(self) -> bool:
        """
        This method will confirm that the user wants to decommission the TiG
        and then update self.unserialize
        """
 
        project_id = self.db.select("SELECT value FROM config WHERE key = 'projectID'")
        job_id = self.db.select("SELECT jobID FROM job")

        try:
            ######################## Confirm that user wants to decommission #######################
            print(f'TiG is commissioned\nProject ID: {project_id}\nJob ID: {job_id}')

            dcm = input('\nDecommission TiG? [y/n] ')
            while dcm not in ['y','Y','n','N']:
                dcm = input('Invalid input. Decommission TiG? [y/n] ')

            if dcm in ['n','N']: return False # User wants to cancel
            ########################################################################################

            ############################ Ask if user wants to unserialize ##########################
            uns = input('Unserialize TiG? [y/n] ')
            while uns not in ['y','Y','n','N']:
                uns = input('Invalid input. Unserialize TiG? [y/n] ')
            ########################################################################################
        except KeyboardInterrupt: return False # CTRL+C to exit
        
        self.unserialize = True if uns in ['y','Y'] else False




if __name__ == '__main__':
    """
    Run this script to commission/decommission TiG

    This script can be run with or without arguments

    Example of running this script with arguments to commission:
    # python3 /home/pi/python-scripts/commission --gwName=tig1234 --projectID=308 --jobID=2847

    Example of running this script with arguments to decommission:
    # python3 /home/pi/python-scripts/commission --unserialize=1
    """

    # Connect to local sqlite database
    db = Database()

    # Check if commissioned
    commissioned = db.select("SELECT value FROM config WHERE key = 'commissioned'")

    if commissioned == 0: Commission(db)
    elif commissioned == 1: Decommission(db)

    db.close()
