from constants import HOST, DATABASE_PATH
from functions import log
import sqlite3


class Database:
    """
    This class connects to the local sqlite database on the TiG
    and is used to execute queries

    There are two methods for executing queries:
    'select' for select queries
    'execute' for all other queries
    """

    def __init__(self):
        self.connect()


    def connect(self):
        self.con = sqlite3.connect(DATABASE_PATH)
        self.cur = self.con.cursor()

    
    def select(self, query:str, fetch_one=True):
        """
        Use this method when selecting from the database

        fetch_one = False will return a list of rows
        fetch_one = True will return the first value of the first row

        This method will return 'None' if fetch_one = True
        but no rows were fetched, or if the query fails for any reason
        """

        try:
            with self.con:
                self.cur.execute(query)
                rows = self.cur.fetchall()

                if fetch_one == False: return rows

                # fetchall() will return a list of tuples
                # Each tuple is a row in the database
                if rows: return rows[0][0]

                # If no rows were fetched: return None
                log(f'Query: {query} | No rows fetched')
                return None
            
        except sqlite3.Error as e:
            log(f'Query failed: {query} | {e}')
            return None


    def execute(self, query:str, parameters=[]):
        """
        Use this method for general (non-select) queries
        such as insert, update, replace or delete

        For select queries, use the 'select' method of this class

        Providing a list of parameters will cause this method
        to use executemany(). Otherwise, it uses execute()

        This method will return the number of rows affected by the query
        or 0 if the query fails for any reason
        """

        try:
            with self.con:
                if parameters == []: self.cur.execute(query)
                else: self.cur.executemany(query, parameters)
                return self.cur.rowcount
        
        except sqlite3.Error as e:
            log(f'Query failed: {query} | {e}')
            return 0
        
    
    def close(self):
        self.con.close()




if __name__ == '__main__':
    """Run this script to initialize database"""
    
    db = Database()

    # Create tables
    for query in [
        """CREATE TABLE IF NOT EXISTS config (
        key TEXT PRIMARY KEY UNIQUE,
        value BLOB)""",

        """CREATE TABLE IF NOT EXISTS gateway (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        gwID INTEGER,
        gwKey TEXT,
        name TEXT UNIQUE,
        modified TEXT)""",

        """CREATE TABLE IF NOT EXISTS job (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        jobID INTEGER UNIQUE,
        name TEXT,
        address TEXT,
        contactName TEXT,
        archived INTEGER,
        created TEXT,
        modified TEXT)""",

        """CREATE TABLE IF NOT EXISTS node (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        nodeID INTEGER UNIQUE,
        phyID INTEGER UNIQUE,
        name TEXT,
        created TEXT,
        modified TEXT)""",

        """CREATE TABLE IF NOT EXISTS sensor (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        sensorID INTEGER UNIQUE,
        name TEXT,
        type INTEGER,
        input INTEGER,
        comments TEXT,
        wsID INTEGER,
        sensorTypeID INTEGER,
        sensorTypeName TEXT,
        offset INTEGER,
        custom1 DOUBLE, custom2 DOUBLE, custom3 DOUBLE,
        created TEXT,
        modified TEXT,
        tempCompSensorID INTEGER,
        phyID INTEGER)""",

        """CREATE TABLE IF NOT EXISTS data (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        node INTEGER,
        input INTEGER,
        ts INTEGER,
        value INTEGER,
        synced INTEGER)""",

        """CREATE TABLE IF NOT EXISTS alarms (
        phyID INTEGER,
        input INTEGER,
        name TEXT UNIQUE,
        onInequality TEXT,
        onValue FLOAT,
        onReadings INTEGER,
        onCount INTEGER,
        offInequality TEXT,
        offValue FLOAT,
        offReadings INTEGER,
        offCount INTEGER,
        relay INTEGER,
        created TEXT,
        UNIQUE (
            phyID,
            input,
            onInequality,
            onValue,
            onReadings,
            offInequality,
            offValue,
            offReadings
        )
        )""",

        """CREATE TABLE IF NOT EXISTS alarmLogs (
        phyID INTEGER,
        input INTEGER,
        name TEXT,
        onDate TEXT,
        offDate TEXT
        )"""

    ]: db.execute(query)

    # Initialize config table
    username, password = 'tig', '20075westb'
    if HOST == 'reproduction':
        username, password = 'nick', 'smt'        

    for row in [
        ('analyticsUsername', username),
        ('analyticsPassword', password),
        ('PHPSESSID',        ''),
        
        ('commissioned',      0),
        ('projectID',         0),
        ('pan',               0),
        ('samplingTimer',   193),
        ('I3node',            0),
        ('A3node',            0),
        
        ('syncTimer',      3600),
        ('syncing',           0),
        ('lastSyncStatus',    0),
        ('lastSyncDate',     ''),
        ('readingsSynced',    0),
        ('clearInterval',    30),

        ('softwareVersion', '2.1.0'),

    ]: db.execute(f'INSERT INTO config VALUES {row}')

    # Initialize gateway table
    db.execute("INSERT INTO gateway VALUES (1, 0, 'gwKey', 'name', '1990-01-01 12:00:00')")

    # Initialize job table
    db.execute("INSERT INTO job VALUES (1, 0, 'name', 'address', 'contactName', 0, '1990-01-01 12:00:00', '1990-01-01 12:00:00')")

    # Create triggers
    db.execute("""CREATE TRIGGER alarmOn
    AFTER UPDATE OF onCount ON alarms
    WHEN NEW.onCount = NEW.onReadings
    BEGIN
        INSERT INTO alarmLogs (phyID, input, name, onDate) VALUES (NEW.phyID, NEW.input, NEW.name, DATETIME('now'));
        SELECT alarm_on(NEW.phyID, NEW.input, NEW.relay);
    END""")

    db.execute("""CREATE TRIGGER alarmOff
    AFTER UPDATE OF offCount ON alarms
    WHEN NEW.offCount = NEW.offReadings
    BEGIN
        UPDATE alarmLogs SET offDate = DATETIME('now') WHERE name = NEW.name AND offDate is null;
        UPDATE alarms SET onCount = 0 WHERE name = NEW.name;
        SELECT alarm_off(NEW.phyID, NEW.input, NEW.relay);
    END""")

    # Change to journal mode to WAL
    db.execute('PRAGMA journal_mode=WAL')

    log('Database initialized')

    db.close()
