from database import Database
from constants import *
from functions import *
import RPi.GPIO as GPIO
import threading
import requests
import time
import zmq
import os


class LED:
    """This class is used to turn an LED on/off, or cause it to blink on and off"""

    def __init__(self, pin, blinks=False):
        self.pin = pin
        self.blinks = blinks

        if self.blinks:
            self.thread = threading.Thread(target=self.wait)
            self.thread.start()


    # Toggle LED on/off
    def toggle(self, state:int): 
        if self.blinks: self.blink = False
        GPIO.output(self.pin, state)


    # If this LED blinks, it continuously waits for self.blink=True
    # Then it will blink until self.blink=False or self.running=False
    def wait(self):
        self.blink = False
        self.running = True

        while self.running:
           time.sleep(0.3)
           if self.blink: GPIO.output(self.pin, (not GPIO.input(self.pin)))




class InterfaceController:
    """
    This is the main class for running the TiG Interface Controller (ifc)
    
    It contains the main infinite while loop that will run until:
        -the user manually ends this script
        -the user manually shuts down / restarts the RPi 
        -the RPi restarts automatically at midnight

    This class controls the LEDs on the TiG case
    and handles all button presses by the user
    """

    def __init__(self):
        # Variable that will remain True until a KeyboardInterrupt (CTRL+C) is received
        # Note: KeyboardInterrupt (CTRL+C) is equivalent to kill -2 <PID of this script>
        self.running = True

        # This script will restart/shutdown the RPi
        # if the power button is pressed/held
        self.restart_rpi = False
        self.shutdown_rpi = False

        # The last time a button was pressed
        self.last_button_edge = 0

        # Communication with ifc_main.py
        self.socket = zmq.Context().socket(zmq.PAIR)
        self.socket.bind(ZMQ_SOCKET)

        # GPIO setup
        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BOARD)

        # Set up button pins
        self.btn_pins = [PWR_BTN_PIN, WIFI_BTN_PIN, SYNC_BTN_PIN]
        for pin in self.btn_pins:
            GPIO.setup(pin, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            GPIO.add_event_detect(pin, GPIO.FALLING, callback=self.button_press)

        # Set up LED pins
        self.led_pins = [STATUS_G_PIN, STATUS_R_PIN, SYNC_G_PIN, SYNC_R_PIN, A_PIN, B_PIN]
        for pin in self.led_pins:
            GPIO.setup(pin, GPIO.OUT, initial=0)

        # LED objects
        self.status_g_led = LED(STATUS_G_PIN)
        self.status_r_led = LED(STATUS_R_PIN)
        self.sync_g_led   = LED(SYNC_G_PIN, blinks=True)
        self.sync_r_led   = LED(SYNC_R_PIN)
        self.a_led        = LED(A_PIN)
        self.b_led        = LED(B_PIN, blinks=True)

        # Thread that waits for messages from bb_main.py
        self.zmq_thread = threading.Thread(target=self.zmq_receive)
    

    def begin(self):

        self.zmq_thread.start()
        self.update_LEDs()

        # Begin infinite while loop
        try:
            while self.running: time.sleep(1)
        # KeyboardInterrupt is the same as kill -2 <PID of tigifc.service> or CTRL+C
        except KeyboardInterrupt: self.running = False


    def end(self):

        log(f'Ending ifc_main.py [PID {current_pid}]...')

        self.zmq_thread.join()

        # Turn off all LEDs
        for led in [self.status_g_led, self.status_r_led, self.sync_g_led, self.sync_r_led, self.a_led, self.b_led]: 
            if led.blinks:
                led.running = False
                led.thread.join()
            led.toggle(0)

        # Clean up GPIO pins
        for pin in self.led_pins+self.btn_pins:
            GPIO.cleanup(pin)

        # Restart/shutdown RPi if user pressed/held power button on TiG interface controller
        if self.restart_rpi:
            log('Restarting RPi...')
            shell_cmd('sudo shutdown -r now')
        elif self.shutdown_rpi:
            log('Shutting down RPi...')
            shell_cmd('sudo shutdown -h now')

    
    def button_press(self, channel):
        """This method is called whenever a button is pressed"""

        if time.time() - self.last_button_edge < 0.5: pass # Button debounce

        elif channel == PWR_BTN_PIN:

            log('Power button on TiG pressed')
            self.b_led.blink = True

            held = 0
            while not GPIO.input(PWR_BTN_PIN) and held < 30:
                time.sleep(0.1); held += 1

            # Restart RPi if power button was held for less than 3 seconds
            # Shutdown RPi if power button was held for 3 or more seconds
            self.running = False
            self.restart_rpi = True if held <= 29 else False
            self.shutdown_rpi = True if held == 30 else False
        
        elif channel == WIFI_BTN_PIN:

            log('WiFi button on TiG pressed')
            
            held = 0
            while not GPIO.input(WIFI_BTN_PIN) and held < 30:
                time.sleep(0.1); held += 1

            # Update LEDs if WiFi button was held for less than 3 seconds
            # Say IP addresses if WiFi button was held for 3 or more seconds
            self.say_ip_addresses() if held == 30 else self.update_LEDs()

        elif channel == SYNC_BTN_PIN:

            log('Sync button on TiG pressed')
            shell_cmd('bigberry sync')

        self.last_button_edge = time.time()


    def update_LEDs(self):

        # Connect to local sqlite database
        db = Database()

        # Turn off all LEDs (except A LED) for 0.3 seconds
        for led in [self.status_g_led, self.status_r_led, self.sync_g_led, self.sync_r_led, self.b_led]: 
            led.toggle(0)
        time.sleep(0.3)

        # Status LED
        if self.internet_on(): self.status_g_led.toggle(1) 
        else: self.status_r_led.toggle(1)

        # Sync LED
        status  = db.select("SELECT value FROM config WHERE key = 'lastSyncStatus'")
        syncing = db.select("SELECT value FROM config WHERE key = 'syncing'")

        if syncing: self.sync_g_led.blink = True
        elif status == 1: self.sync_g_led.toggle(1)
        elif status == 0: self.sync_r_led.toggle(1)

        # B LED
        if get_service_pid('bigberry.service') > 0: self.b_led.toggle(1)

        db.close()


    def internet_on(self) -> bool:
        """Returns True if response code from Google = 200"""

        try:
            response = requests.get('http://www.google.com')
            return response.status_code == 200
        except requests.exceptions.RequestException: return False


    def say_ip_addresses(self):
        """
        This method is called if the WiFi button is held for 3 seconds
        It will say the ethernet/WiFi/USB IP addresses through the RPi headphone jack
        """

        l = ['zero','one','two','three','four','five','six','seven','eight','nine']
        d = {str(i): word for i, word in enumerate(l)}
        d.update({'.': 'dot'})

        text = ''

        for string, connection in [
            ('The ethernet I_P_address_is ', 'eth0' ),
            ('The wireless I_P_address_is ', 'wlan0'),
            ('The U_S_B I_P_address_is ',    'usb0' )
        ]:
            addr = get_ip_address(connection)
            if addr: text += string
            for char in addr: text += f'{d[char]} '

        try:
            for word in text.split(): shell_cmd(f'espeak {word} 2>/dev/null')
        except KeyboardInterrupt: pass


    # Thread
    def zmq_receive(self):
        """
        This thread waits for a message from bb_main.py
        Messages control the LEDs
        """

        while self.running:

            # Wait 1000ms for a message from ifc_main.py
            if not self.socket.poll(timeout=1000): continue

            cmd = self.socket.recv_pyobj()

            # Syncing to Analytics
            if cmd == SYNC_LED_BLINK_GREEN:
                self.sync_r_led.toggle(0)
                self.sync_g_led.blink = True

                # Update status LED every sync
                if self.internet_on():
                    self.status_r_led.toggle(0)
                    self.status_g_led.toggle(1)
                else: 
                    self.status_g_led.toggle(0)
                    self.status_r_led.toggle(1)

            # Sync to Analytics failed
            elif cmd == SYNC_LED_RED:
                self.sync_g_led.toggle(0)
                self.sync_r_led.toggle(1)

            # Sync to Analytics passed
            elif cmd == SYNC_LED_GREEN:
                self.sync_r_led.toggle(0)
                self.sync_g_led.toggle(1)

            # Inserting data into data table
            elif cmd == A_LED_ON:
                self.a_led.toggle(1)

            # Not inserting data into data table
            elif cmd == A_LED_OFF:
                self.a_led.toggle(0)

            # BiGBerry is running
            elif cmd == B_LED_ON:  
                self.b_led.toggle(1)

            # BiGBerry is not running
            elif cmd == B_LED_OFF:
                self.b_led.toggle(0)




if __name__ == '__main__':
    """
    It is important that only one instance of this script is running at a time
    This is checked by comparing the PID of this script with the PID of tigifc.service

    If this script was run by starting tigifc.serivce, the two PIDs will be the same,
    and this script will start the TiG Interface Controller

    Otherwise, the two PIDs will be different, and this script will do nothing

    tigifc.service can only be started once, so running this script by starting tigifc.service
    will ensure that only one instance of this script can ever be running at a time

    However, this means that the ONLY way to run this script is by starting tigifc.service
    To bypass this, remove the if-else statement below
    """
    
    current_pid = os.getpid()
    service_pid = get_service_pid('tigifc.service')

    log(f'############## RUNNING ifc_main.py [PID {current_pid}] ##############')

    if current_pid == service_pid:
        tigifc = InterfaceController()
        tigifc.begin()
        # Wait for KeyboardInterrupt (CTRL+C) or kill -2 <PID of this script>
        tigifc.end()
    else:
        log(f'ERROR! ifc_main.py PID does not match tigifc.service PID [{service_pid}]')

    log(f'############## END OF ifc_main.py [PID {current_pid}] ##############')
