from datetime import datetime, timezone
from usbhub import USBHub
from relay import Relay
from constants import *
from functions import *
from api import API
import threading
import math
import time


class Sync:
    """
    This class is responsible for uploading unsynced data
    from the local database on the TiG
    to the SMT Analytics server database

    An instance of this class is created in bb_main.py
    When commissioning, an instance of this class is created in commission.py
    """

    def __init__(self, usb:USBHub, socket):
        # For logging
        self.usb = usb

        # For sending messages to ifc_main.py
        self.socket = socket
        
        # Thread that waits until it has been enough time since last sync
        self.thread = threading.Thread(target=self.wait)
        self.running = True

    
    # Thread
    def wait(self):
        # Connect to local sqlite database
        # This can only done inside this thread
        # Relay class inherits from Database class
        db = Relay()

        # For making API calls
        api = API(db)

        # How often TiG syncs to Analytics (seconds)
        self.sync_timer = self.get_sync_timer(db)

        # When the last sync took place (epoch time)
        self.last_sync_time = time.time()

        while self.running:
            if time.time() - self.last_sync_time > self.sync_timer: self.sync_data(db, api)
            else: time.sleep(1) # Keep waiting

        db.close()


    def force_sync(self):
        self.last_sync_time = 0


    def get_sync_timer(self, db:Relay) -> int:

        timer = db.select("SELECT value FROM config WHERE key = 'syncTimer'")

        # Allowed values for sync timer are between 10 minutes and 23 hours
        if 600 <= timer <= 23*3600: return timer

        # If for some reason the value in the database is outside of this range,
        # set it to the default value of one hour
        db.execute("UPDATE config SET value = 3600 WHERE key = 'syncTimer'")
        return 3600


    def sync_data(self, db:Relay, api:API):

        log(f'################# Syncing data with {HOST} #################')
        # Update config table in database
        db.execute("UPDATE config SET value = 1 WHERE key = 'syncing'")


        if self.usb is not None: self.usb.log_usb_info(reset_counts=True)
        # Sync LED starts blinking
        if self.socket is not None: self.socket.send_pyobj(SYNC_LED_BLINK_GREEN)

        # Gateway ID
        self.gw_id = db.select("SELECT gwID FROM gateway")
        # How many rows need to be synced (for logging)
        self.num_unsynced = db.select('SELECT count(id) FROM data WHERE synced = 0')
        # How many rows were actually synced (for logging)
        self.num_synced = 0
        # Result of this sync
        self.sync_passed = 0
        

        # Don't proceed with sync if TiG is not commissioned
        commissioned = db.select("SELECT value FROM config WHERE key = 'commissioned'")
        if commissioned == 0:
            log('TiG not commissioned')
            self.finish_sync(db)
        

        self.insert_diagnostics(db)
        self.insert_distinct_nodes_sensors(db)


        # Don't proceed with sync if exchange schema fails
        if api.exchange_schema() == False:
            log('Exchange schema failed')
            self.finish_sync(db)

        
        # Row ID (min and max) of unsynced rows in data table
        min_id, max_id = db.select('SELECT min(id), max(id) FROM data WHERE synced = 0', fetch_one=False)[0]
        # If there are no unsynced rows in data table, min_id and max_id will be None
        if min_id is None or max_id is None: min_id = max_id = 0
        # Number of chunks for this sync (for logging)
        num_chunks = math.ceil( (max_id-min_id)/SYNC_CHUNK_SIZE )
        # Current chunk (for logging)
        curr_chunk = 1
        # Assume this sync will pass
        self.sync_passed = 1


        # Loop through rows in database (in chunks)
        for lower_id in range(min_id, max_id, SYNC_CHUNK_SIZE):

            # CSV string that will be written to /home/pi/csv-data/data.csv
            csv_string = ''
            # Number of lines in CSV string
            csv_count = 0

            # Fetch rows from database
            upper_id = min(max_id, lower_id+SYNC_CHUNK_SIZE)
            rows = db.select(f'SELECT * FROM data WHERE id >= {lower_id} AND id <= {upper_id} AND synced = 0', fetch_one=False)

            log(f'Uploading {len(rows)} data entries to {HOST} [chunk {curr_chunk}/{num_chunks}]...')
            
            # Loop through rows from previous 'SELECT' query
            for row_id, phy_id, sensor_input, ts, value, synced in rows:
                # Get sensor ID
                sensor_id = db.select(f'SELECT sensorID FROM sensor WHERE phyID = {phy_id} AND input = {sensor_input}')
                if sensor_id:
                    utc_ts = datetime.fromtimestamp(ts, tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
                    csv_string += f'{sensor_id},{utc_ts},{value}\n'
                    csv_count += 1
                # If query to get sensor ID returned 'None'
                else: self.sync_passed = 0

            # Upload data
            if api.upload_data(csv_string):
                query = f'UPDATE data SET synced = 1 WHERE id >= {lower_id} AND id <= {upper_id} AND synced = 0'
                rows_updated = db.execute(query)
                self.num_synced += csv_count
            else:
                self.sync_passed = 0
                rows_updated = 0

            log(f'{rows_updated}/{len(rows)} rows in data table updated [chunk {curr_chunk}/{num_chunks}]')
            curr_chunk += 1


        self.finish_sync(db)


    def insert_diagnostics(self, db:Relay):

        log('Inserting diagnostics...')

        for i, val in enumerate([
            # Input 1 -> Storage space used (slice to remove percent sign)
            shell_cmd("df -hl | grep root | awk '{print $5}'")[:-1],

            # Input 2 -> Uptime (slice to remove decimal)
            shell_cmd("cat /proc/uptime | awk '{print $1}'")[:-3],

            # Input 3 -> Number of unsynced rows in data table
            self.num_unsynced,

            # Input 4 -> Public IP address
            ip_address_to_int(shell_cmd('curl ifconfig.me/ip')),

            # Input 5 -> Local IP address (eth0, wlan0, or usb0)
            ip_address_to_int(get_ip_address_priority()),

            # Input 6 -> WiFi connection
            1 if get_ip_address('wlan0') else 0
        ]):
            row = (self.gw_id, i+1, int(time.time()), int(val), 0)
            self.num_unsynced += db.execute(f"INSERT INTO data (node, input, ts, value, synced) VALUES {row}")


    def insert_distinct_nodes_sensors(self, db:Relay):

        log('Inserting distinct nodes and sensors...')
        
        nodes_to_insert = nodes_inserted = 0
        sensors_to_insert = sensors_inserted = 0
        utc_time = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")

        # Sensor names for inputs 1 to 6
        sensor_names = {
            1: 'Storage used',
            2: 'Uptime',
            3: 'Unsynced readings',
            4: 'Public IP',
            5: 'Local IP',
            6: 'WiFi connection'
        }

        for sensor_input, node in db.select('SELECT distinct input, node FROM data', fetch_one=False):
            
            in_node_table   = db.select(f"SELECT 1 FROM node   WHERE phyID = {node}")
            in_sensor_table = db.select(f"SELECT 1 FROM sensor WHERE phyID = {node} AND input = {sensor_input}")

            if in_node_table != 1: # Not in node table
                nodes_to_insert += 1
                nodes_inserted += db.execute(f"""INSERT INTO node
                (phyID, name, created, modified) VALUES
                {(node, 'New SMT-A2', utc_time, utc_time)}""")

            if in_sensor_table != 1: # Not in sensor table
                name = sensor_names[sensor_input] if node == self.gw_id else 'New'

                sensors_to_insert += 1
                sensors_inserted += db.execute(f"""INSERT INTO sensor
                (name, type, input, wsID, sensorTypeID, sensorTypeName, offset,
                custom1, custom2, custom3, created, modified, tempCompSensorID, phyID) VALUES
                {(name, 0, sensor_input, -1, -1, 'Unknown', 0, 0, 0, 0, utc_time, utc_time, -1, node)}""")

        log(f'{nodes_inserted}/{nodes_to_insert} distinct nodes and {sensors_inserted}/{sensors_to_insert} distinct sensors inserted into database')


    def finish_sync(self, db:Relay):

        # Update config table in database
        db.execute(f"UPDATE config SET value = {self.sync_passed}    WHERE key = 'lastSyncStatus'")
        db.execute(f"UPDATE config SET value = '{shell_cmd('date')}' WHERE key = 'lastSyncDate'"  )
        db.execute(f"UPDATE config SET value = {self.num_synced}     WHERE key = 'readingsSynced'")
        db.execute( "UPDATE config SET value = 0                     WHERE key = 'syncing'"       )

        self.delete_old_data(db)
        self.last_sync_time = time.time()
        self.sync_timer = self.get_sync_timer(db)

        # Update sync LED on TiG interface controller
        if self.socket is not None:
            if self.sync_passed: self.socket.send_pyobj(SYNC_LED_GREEN)
            else: self.socket.send_pyobj(SYNC_LED_RED)

        log_msg = 'Sync passed' if self.sync_passed else 'Sync failed'
        log(log_msg+f' | Readings synced: {self.num_synced} out of {self.num_unsynced}')

        log('###################### Sync complete ######################')
        

    def delete_old_data(self, db:Relay):

        cutoff = db.select("SELECT value FROM config WHERE key = 'clearInterval'")
        if cutoff < -1: return

        # 'clearInterval' value in config table shouldn't be less than 7 (days)
        # If for some reason it is, reset it to its default value of 30 (days)
        if cutoff < 7:
            db.execute("UPDATE config SET value = 30 WHERE key = 'clearInterval'")
            cutoff = 30

        cutoff_ts = int(time.time() - cutoff*24*3600)
        deleted = db.execute(f"DELETE FROM data WHERE ts < {cutoff_ts} AND synced = 1")
        log(f'Deleted {deleted} synced rows in data table older than {cutoff} days')
