from database import Database
from wimote import Wimote
from cankey import CANKey
from functions import log
from constants import *
from relay import Relay
import threading
import usb.core
import usb.util
import queue
import time


class USBHub:
    """
    This class is responsible for keeping an updated list
    of Wimote and CANKey objects representing
    SMT USB devices connected to the TiG

    Each Wimote and CANKey object created by this class
    puts received sensor data in self.data_queue

    The wait() thread of this class will wait for data
    to be put in self.data_queue, and then call the
    save_data() method of this class, which will
    store the data in the local database on the TiG

    An instance of this class is created in bb_main.py
    """

    def __init__(self, socket):
        # For sending messages to ifc_main.py
        self.socket = socket

        # All Wimote and CANKey objects put data in self.data_queue
        self.data_queue = queue.Queue()

        # Thread that waits until there is data in self.data_queue
        self.thread = threading.Thread(target=self.wait)
        self.running = True

        self.wimotes:list[Wimote] = []
        self.cankeys:list[CANKey] = []

        # Connect to local database
        self.db = Database()


    # Thread
    def wait(self):
        # Connect to local sqlite database
        # This can only done inside this thread
        # Relay class inherits from Database class
        db = Relay()

        # Check for any alarms that should be triggered
        db.check_alarms()

        while self.running:
            if self.data_queue.empty(): time.sleep(1) # Keep waiting
            else: self.save_data(db)

        db.close()

    
    def save_data(self, db:Relay):
        
        # Turn on A LED
        self.socket.send_pyobj(A_LED_ON)

        # While loop that runs until self.data_queue is empty
        # Needed in case BiGBerry tries to shutdown while there are messages in queue
        while not self.data_queue.empty():
            data = []

            # While loop that runs until DATA_CHUNK_SIZE messages are read
            # Needed in case RPi loses power while there are messages in queue
            while len(data) < DATA_CHUNK_SIZE and not self.data_queue.empty():
                data.append(self.data_queue.get_nowait())

            # Insert chunk of data into database
            if data:
                query = "INSERT INTO data (node, input, ts, value, synced) VALUES (?,?,?,?,?)"
                db.execute(query, parameters=data)

        # Turn off A LED when queue is empty
        self.socket.send_pyobj(A_LED_OFF)

    
    def update_usb_lists(self):
        """
        This method is called every time a SMT USB device is plugged into/unplugged from the RPi
        (It is also called every hour by crontab)

        Its purpose is to check self.wimotes and self.cankeys for any Wimote or CANKey objects
        that need to be added to/removed from their respective lists

        The file /etc/udev/rules.d/88-bigberry.rules
        will detect when a SMT USB device is plugged into/unplugged from the RPi

        Then, it will then execute the command "bigberry updateUSB"
        which is equivalent to kill -7 <PID of bigberry.service>

        The 'signal_handler' method in bb_main.py will then receive the kill signal,
        and call this method
        """
        
        time.sleep(1)

        # List of addresses (unique integers) of all devices currently connected to the RPi
        con_addrs = []
        # List of addresses (unique integers) of devices currently in self.wimotes 
        wim_addrs = [wimote.address for wimote in self.wimotes]
        # List of addresses (unique integers) of devices currently in self.cankeys
        can_addrs = [cankey.address for cankey in self.cankeys]

        # Check for new SMT USB devices that have been plugged into the RPi
        # and add them to their proper list
        for dev in usb.core.find(find_all=True, idVendor=0x04d8):

            con_addrs.append(dev.address)

            if dev.idProduct == 0xfd70 and dev.address not in wim_addrs:
                self.wimotes.append(Wimote(dev, self.data_queue))

            elif dev.idProduct == 0xff0f and dev.address not in can_addrs:
                self.cankeys.append(CANKey(dev, self.data_queue))

        # Check for wimotes that have been unplugged from RPi
        # and remove them from self.wimotes
        i = 0
        while i < len(self.wimotes):
            wimote = self.wimotes[i]
            if wimote.address in con_addrs: i += 1 # Still connected to RPi
            else: wimote.stop(); del self.wimotes[i]
    
        # Check for cankeys that have been unplugged from RPi
        # and remove them from self.cankeys
        i = 0
        while i < len(self.cankeys):
            cankey = self.cankeys[i]
            if cankey.address in con_addrs: i += 1 # Still connected to RPi
            else: cankey.stop(); del self.cankeys[i]


    def wimote_write(self, kill_signal:int):

        cmd, key = {
            3: (USB_SET_RTC,    'samplingTimer'),
            4: (USB_SET_PAN,    'pan'          ),
            5: (USB_RETRANSMIT, 'A3node'       ),
            6: (USB_ERASE,      'A3node'       )
        }[kill_signal]

        value   = self.db.select(f"SELECT value FROM config WHERE key = '{key}'")
        I3_node = self.db.select( "SELECT value FROM config WHERE key = 'I3node'")
        
        for wimote in self.wimotes:
            # The only wimote that will write this command is the one
            # whose node_id matches the I3_node value in the config table
            if wimote.node_id == I3_node: wimote.write(cmd, value); break
        else: log(f'WARNING: No Wimote connected with id {I3_node}')


    def write_usb_info(self):
        """Write info of USB devices to /home/pi/device-info/dev_info.txt"""

        dev_info = ''

        # Make sure attributes are separated with "|" and lines are separated with "\n" (for RaspAp web UI)
        for wimote in self.wimotes:
            timer_string = wimote.timer_to_string(wimote.sampling_timer)
            dev_info += f'Wimote|{wimote.node_id}|{wimote.pan}|{timer_string}|{wimote.running}\n'
        
        for cankey in self.cankeys:
            dev_info += f'CANKey|{cankey.address}|{cankey.running}\n'

        try:
            with open(DEV_INFO_PATH, 'w') as f: f.write(dev_info)
        except Exception as e:
            log(f'Unable to open {DEV_INFO_PATH} | {e}')


    def log_usb_info(self, reset_counts=False):
        for dev in self.wimotes + self.cankeys: dev.log(reset_counts)


    def stop_devices(self):
        for dev in self.wimotes + self.cankeys: dev.stop()
