from datetime import datetime, timezone
from functions import log
from constants import *
import threading
import usb.core
import usb.util
import queue
import time


class Wimote:
    """
    Instances of this class are created in the 'update_usb_lists' method
    of the USBHub class in usbhub.py

    The list 'wimotes' in USBHub will contain one instance of this class
    for each wimote connected to the TiG

    The 'read' and 'write' methods of this class are 
    from the perspective of the TiG

    For example, the 'write' method will send messages
    from the TiG to the wimote

    The 'read' method (which is a thread) will wait for messages sent
    from the wimote to the TiG
    """

    def __init__(self, dev, data_queue:queue.Queue):

        # pyUSB device object
        self.dev = dev

        if self.dev.is_kernel_driver_active(0): 
            self.dev.detach_kernel_driver(0)

        self.address = dev.address # integer (unique)
        self.idProduct = dev.idProduct # 0xfd70
        self.ep_in, self.ep_out = dev[0][(0,0)]

        # Data received by this wimote object gets put into this queue
        # This is the same queue used by all Wimote and CANKey objects
        # in the 'wimote' and 'cankey' lists in USBHub
        self.data_queue = data_queue

        # Dictionary of A3's connected to this Wimote
        # key: node ID of A3
        # value: list of four numbers:
            # index 0 -> Time USB_START_LOG message was received
            # index 1 -> Log size
            # index 2 -> Number of readings received for this log
            # index 3 -> Number of readings received since last sync
        self.nodes = {}

        self.node_id = self.pan = self.sampling_timer = self.rtc = 0

        # Thread that waits for incoming messages from this wimote device
        self.thread = threading.Thread(target=self.read)
        self.running = True
        self.thread.start()

        log(f'[Wimote address {self.address}] - CONNECTED')
        self.setup()


    def setup(self):

        # Get node ID
        # self.node_id is set when 'USB_GET_ID' message is received
        self.write(USB_GET_ID)
        while self.node_id == 0 and self.running: time.sleep(1)

        if not self.running: return

        # Get PAN
        # self.pan is set when 'USB_GET_PAN' message is received
        self.write(USB_GET_PAN)
        while self.pan == 0 and self.running: time.sleep(1)

        if not self.running: return

        # Get sampling timer
        # self.sampling_timer is set when 'USB_GET_RTC' message is received
        self.write(USB_GET_RTC)
        while self.sampling_timer == 0 and self.running: time.sleep(1)

        if not self.running: return

        # Set real-time clock and sampling timer
        # self.rtc is set when 'USB_SET_RTC' message is received
        self.write(USB_SET_RTC, self.sampling_timer)
        while self.rtc == 0 and self.running: time.sleep(1)

        if not self.running: return

        # This message will cause the green LED on the wimote to start blinking
        self.write(USB_ENABLE_CH)

        if not self.running: return

        log(f'[Wimote {self.node_id}] - SETUP COMPLETE')


    # Thread
    def read(self):

        while self.running:

            # Try to receive a message from this wimote device
            # If there are no messages, this will raise a timeout error
            try: rx = self.dev.read(self.ep_in.bEndpointAddress, 64)
            except usb.core.USBError as e:
                if e.args[0] == 110: continue # Ignore timeout error
                else: self.stop(err_msg=f'READING ERROR: {e}')

            #
            # Message received -> proceed
            #

            # First number in message is the command
            cmd = rx[0]

            if cmd == USB_PING_SRC: self.write(USB_ACK_PING)
            
            elif cmd == USB_GET_ID:
                # 'USB_GET_ID' messages are received from wimote
                # in repsonse to 'USB_GET_ID' messages being sent to wimote

                self.node_id = (rx[1] & 0xff) << 24 | (rx[2] & 0xff) << 16 | (rx[3] & 0xff) << 8 | (rx[4] & 0xff)
                log(f'[Wimote address {self.address}] - Node ID: {self.node_id}')
  
            elif cmd == USB_GET_PAN:
                # 'USB_GET_PAN' messages are received from wimote
                # in repsonse to 'USB_GET_PAN' messages being sent to wimote

                self.pan = (rx[6] & 0xff) << 8 | (rx[7] & 0xff)
                log(f'[Wimote {self.node_id}] - PAN: {self.pan}')
    
            elif cmd == USB_SET_PAN:
                # 'USB_SET_PAN' messages are received from wimote
                # in repsonse to 'USB_SET_PAN' messages being sent to wimote

                self.pan = (rx[6] & 0xff) << 8 | (rx[7] & 0xff)
                log(f'[Wimote {self.node_id}] - PAN set to {self.pan}')
          
            elif cmd == USB_GET_RTC:
                # 'USB_GET_RTC' messages are received from wimote
                # in repsonse to 'USB_GET_RTC' messages being sent to wimote

                self.sampling_timer = rx[6]
                timer_string = self.timer_to_string(self.sampling_timer)
                log(f'[Wimote {self.node_id}] - Sampling timer: {timer_string}')

            elif cmd == USB_SET_RTC:
                # 'USB_SET_RTC' messages are received from wimote
                # in repsonse to 'USB_SET_RTC' messages being sent to wimote

                self.sampling_timer = rx[6]
                timer_string = self.timer_to_string(self.sampling_timer)

                # Calculate timestamp
                timestamp = (rx[7] & 0xff) << 24 | (rx[8] & 0xff) << 16 | (rx[9] & 0xff) << 8 | (rx[10] & 0xff)
                year   = (timestamp >> 26) & 0x3f # Last two digits of year (example: 24 for the year 2024)
                month  = (timestamp >> 22) & 0x0f
                day    = (timestamp >> 17) & 0x1f
                hour   = (timestamp >> 12) & 0x1f
                minute = (timestamp >>  6) & 0x3f
                second = (timestamp >>  0) & 0x3f
                
                self.rtc = datetime(year+2000, month, day, hour, minute, second)
                log(f'[Wimote {self.node_id}] - RTC set to {self.rtc} - Sampling timer set to {timer_string}')

            elif cmd == USB_START_LOG:
                # 'USB_START_LOG' messages are sent from an A3 connected to this wimote

                node_id = (rx[1] & 0xff) << 24 | (rx[2] & 0xff) << 16 | (rx[3] & 0xff) << 8 | (rx[4] & 0xff)
                if node_id >= 65535: continue # Ignore messages with node_id >= 65535

                # If this is the first message received from this A3, add it to self.nodes
                if node_id not in self.nodes:
                    self.nodes[node_id] = [0, 0, 0, 0]
                    log(f'[Wimote {self.node_id}] - Receiving readings from node {node_id}')

                self.nodes[node_id][0] = time.time() # Start time
                self.nodes[node_id][1] = size = (rx[6] << 8) | (rx[7] & 0xff) # Log size
                self.nodes[node_id][2] = 0 # Number of readings received for this log  
                log(f'[Wimote {self.node_id} - node {node_id}] - LOG START - Size: {size}')

            elif cmd == USB_END_LOG:
                # 'USB_END_LOG' messages are sent from an A3 connected to this wimote

                node_id = (rx[1] & 0xff) << 24 | (rx[2] & 0xff) << 16 | (rx[3] & 0xff) << 8 | (rx[4] & 0xff)
                if node_id >= 65535: continue # Ignore messages with node_id >= 65535

                # If this is the first message received from this A3, add it to self.nodes
                if node_id not in self.nodes:
                    self.nodes[node_id] = [0, 0, 0, 0]
                    log(f'[Wimote {self.node_id}] - Receiving readings from node {node_id}')

                # If 'USB_END_LOG' message was received without first receiving 'USB_START_LOG' message
                if self.nodes[node_id][0] == 0:
                    log(f'[Wimote {self.node_id} - node {node_id}] - LOG END message received without LOG START message')
                else:
                    seconds = round(time.time()-self.nodes[node_id][0], 3)
                    size    = self.nodes[node_id][1]
                    count   = self.nodes[node_id][2]
                    log(f'[Wimote {self.node_id} - node {node_id}] - LOG END - Received {count}/{size} in {seconds} seconds')

                self.nodes[node_id][0] = 0

            elif cmd == USB_DATA_SET:
                # 'USB_DATA_SET' messages are sent from an A3 connected to this wimote
                # It indicates that the A3 is sending sensor data

                node_id = (rx[1] & 0xff) << 24 | (rx[2] & 0xff) << 16 | (rx[3] & 0xff) << 8 | (rx[4] & 0xff)
                if node_id >= 65535: continue # Ignore messages with node_id >= 65535

                # If this is the first message received from this A3, add it to self.nodes
                if node_id not in self.nodes:
                    self.nodes[node_id] = [0, 0, 0, 0]
                    log(f'[Wimote {self.node_id}] - Receiving readings from node {node_id}')

                # Extract date/time info from message
                timestamp = (rx[7] & 0xff) << 24 | (rx[8] & 0xff) << 16 | (rx[9] & 0xff) << 8 | (rx[10] & 0xff)
                year   = (timestamp >> 26) & 0x3f # Last two digits of year (example: 24 for the year 2024)
                month  = (timestamp >> 22) & 0x0f
                day    = (timestamp >> 17) & 0x1f
                hour   = (timestamp >> 12) & 0x1f
                minute = (timestamp >>  6) & 0x3f
                second = (timestamp >>  0) & 0x3f
                
                # Date/time received from A3 is in UTC time - it is saved in database in UTC time
                dt = datetime(year+2000, month, day, hour, minute, second, tzinfo=timezone.utc)
                ts = int(datetime.timestamp(dt))

                # Loop through sensors
                for i in range(rx[11]):
                    sensor_val = (rx[13+(i*4)] & 0x7f) << 16 | (rx[14+(i*4)] & 0xff) << 8 | (rx[15+(i*4)] & 0xff)
                    if (rx[13+(i*4)] & 0x80) != 0: sensor_val = sensor_val << 9

                    sensor_input = rx[12+(i*4)] + 1
                    if sensor_input < 256:
                        self.data_queue.put_nowait((node_id, sensor_input, ts, sensor_val, 0))
                        self.nodes[node_id][2] += 1 # Number of readings received for this log
                        self.nodes[node_id][3] += 1 # Number of readings received since last sync
    
            else: log(f'[Wimote {self.node_id}] - Message received - Command: {cmd}')


    def write(self, cmd=0, val=0):

        # Message to be sent to wimote (a list of 64 numbers)
        # First number in message is the command
        tx = [cmd] + [0]*63

        if cmd == USB_SET_RTC:
            # Set the real-time clock (RTC) of this wimote

            # 'USB_SET_RTC' messages also set the sampling timer (ST), so the value
            # of the desired sampling timer needs to be known when setting the RTC

            # This message will also change the RTC and ST of A3s connected to this wimote

            timer_string = self.timer_to_string(val)
            utc_time = datetime.now(timezone.utc).strftime("%y-%m-%d %H:%M:%S")
            log(f'[Wimote {self.node_id}] - Setting RTC to UTC time {utc_time}...')
            log(f'[Wimote {self.node_id}] - Setting sampling timer to {timer_string}...')

            # Calculate timestamp
            dates  = datetime.now(timezone.utc).strftime("%y %m %d %H %M %S").split()
            masks  = [0x3f, 0x0f, 0x1f, 0x1f, 0x3f, 0x3f]
            shifts = [26, 22, 17, 12, 6, 0]
            timestamp = 0

            for date, mask, shift in zip(dates, masks, shifts): 
                timestamp &= ~(mask << shift)
                timestamp |= (int(date) & mask) << shift

            tx[6]  = val
            tx[7]  = (timestamp >> 24) & 0xff
            tx[8]  = (timestamp >> 16) & 0xff
            tx[9]  = (timestamp >>  8) & 0xff
            tx[10] = (timestamp >>  0) & 0xff

        elif cmd == USB_SET_PAN:
            # Set the PAN of this wimote
            # This will NOT change the PAN of A3s connected to this wimote

            log(f'[Wimote {self.node_id}] - Setting pan to {val}...')
            tx[6] = (val >> 8) & 0xff
            tx[7] = (val >> 0) & 0xff

        elif cmd == USB_RETRANSMIT:
            # Retransmit the logs of a specific A3 connected to this wimote
            # the node ID of the A3 is 'val'

            log(f'[Wimote {self.node_id}] - Retransmitting logs for node {val}...')
            tx[6] = (val >> 8) & 0xff
            tx[7] = (val >> 0) & 0xff
            
        elif cmd == USB_ERASE:
            # Erase the logs of a specific A3 connected to this wimote
            # the node ID of the A3 is 'val'

            log(f'[Wimote {self.node_id}] - Erasing logs for node {val}...')
            tx[6] = (val >> 8) & 0xff
            tx[7] = (val >> 0) & 0xff

        try: self.ep_out.write(bytes(tx))
        except usb.core.USBError as e:
            self.stop(err_msg=f'WRITING ERROR: {e}')


    def log(self, reset_counts=False):

        if not self.running:
            log(f'[Wimote {self.node_id}] - STOPPED RUNNING AT {self.stop_time}')
            return

        for node_id in self.nodes:
            count = self.nodes[node_id][3] # Number of readings received since last sync
            if count == 0: continue

            log(f'[Wimote {self.node_id} - node {node_id}] - Readings since last sync: {count}')
            if reset_counts: self.nodes[node_id][3] = 0


    def stop(self, err_msg=''):

        if self.running == False: return

        msg = f'[Wimote {self.node_id}] - STOPPING'
        if err_msg: msg += f' | {err_msg}'
        log(msg)

        self.log()
        self.running = False
        self.stop_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # This message will cause the green LED on the wimote to stop blinking
        self.write(USB_DISABLE_CH)


    def timer_to_string(self, timer:int) -> str:
        """
        Converts sampling timer integer to human-readable string
        example: 193 -> '1 hour(s)'
        """

        r = timer >> 6
        v = timer & 0x3f

        if r == 1: return f'{v*10} second(s)'
        if r == 2: return f'{v} minute(s)'
        if r == 3: return f'{v} hour(s)'
        return f'{timer} [?]'
